package org.eclipse.xtext.example.domainmodel.jvmmodel

import com.google.inject.Inject
import org.eclipse.xtext.example.domainmodel.domainmodel.Entity
import org.eclipse.xtext.example.domainmodel.domainmodel.Operation
import org.eclipse.xtext.example.domainmodel.domainmodel.Property
import org.eclipse.xtext.naming.IQualifiedNameProvider
import org.eclipse.xtext.xbase.jvmmodel.AbstractModelInferrer
import org.eclipse.xtext.xbase.jvmmodel.IJvmDeclaredTypeAcceptor
import org.eclipse.xtext.xbase.jvmmodel.JvmTypesBuilder
import org.eclipse.xtext.xbase.lib.Procedures$Procedure1
import org.eclipse.xtext.common.types.JvmField
import org.eclipse.xtext.example.domainmodel.domainmodel.XmlGenerator
import org.eclipse.xtext.xbase.compiler.TypeReferenceSerializer
import org.eclipse.xtext.xbase.compiler.output.ITreeAppendable
import org.eclipse.xtext.example.domainmodel.domainmodel.XmlTags
import org.eclipse.xtext.example.domainmodel.domainmodel.XmlContent

class DomainmodelJvmModelInferrer extends AbstractModelInferrer {
	
	@Inject extension TypeReferenceSerializer
	@Inject extension JvmTypesBuilder
	@Inject extension IQualifiedNameProvider

	def dispatch infer(Entity entity, IJvmDeclaredTypeAcceptor acceptor, boolean prelinkingPhase) {
		acceptor.accept(
			entity.toClass( entity.fullyQualifiedName )
		).initializeLater [
			documentation = entity.documentation
			if (entity.superType != null)
				superTypes += entity.superType.cloneWithProxies
			val procedure = entity.newTypeRef(typeof(Procedure1), it.newTypeRef())
			members += entity.toConstructor() []
			members += entity.toConstructor() [
				parameters += entity.toParameter("initializer", procedure)
				body = [it.append("initializer.apply(this);")]
			]
			val fields = <JvmField>newArrayList()
			for ( f : entity.features ) {
				switch f {
			
					Property : {
						val field = f.toField(f.name, f.type)
						fields += field
						members += field
						members += f.toGetter(f.name, f.type)
						members += f.toSetter(f.name, f.type)
					}
			
					Operation : {
						members += f.toMethod(f.name, f.type) [
							documentation = f.documentation
							for (p : f.params) {
								parameters += p.toParameter(p.name, p.parameterType)
							}
							body = f.body
						]
					}
					
					XmlGenerator : {
						members += f.toMethod(f.name + "XmlGenerator", f.newTypeRef(typeof(String))) [
							documentation = f.documentation
							body = [
								val varName = declareVariable(this, "sb")
								var t = trace(f, true)
								f.newTypeRef(typeof(StringBuilder)).serialize(f, t)
								t.append(''' varName = new ''')
								f.newTypeRef(typeof(StringBuilder)).serialize(f, t)
								t.append(''' ();''').newLine
								inferXml(f.block, 0)
								trace(f).append('''return varName.toString();''').newLine
							]
						]
					}
				}
			}
			members += entity.toToStringMethod(it)
		]
	}
	
	def dispatch inferXml(ITreeAppendable it, XmlTags tags, int indentation) {
		for (tag : tags.tags) {
			var t =trace(tag, true)
			t.indent(indentation)
			t.outputln('''"<tag.name>"''')
			inferXml(tag.block, indentation + 1)
			trace(tag).indent(indentation)
			trace(tag).outputln('''"</tag.name>"''')
		}
	}
	def dispatch void inferXml(ITreeAppendable it, XmlContent content, int indentation) {
		val t = trace(content, true)
		t.indent(indentation)
		t.outputln('''this.content.property.name''')
	}
	
	def indent(ITreeAppendable it, int indentation) {
		val spaces = (0..indentation).map [""].reduce [ result, next |
			result + "  "
		]
		output('''"spaces"''')
	}
	
	def outputln(ITreeAppendable it, String content) {
		val varName = getName(this)
		append('''varName.append(content);''').newLine
		append('''varName.append("\n");''').newLine
		append('''System.out.println(content);''').newLine
	}
	
	def output(ITreeAppendable it, String content) {
		val varName = getName(this)
		append('''varName.append(content);''').newLine
		append('''System.out.print(content);''').newLine
	}
}